# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import numpy as np
import cv2

from . import data_path
from ..dataset import Dataset, Split


class WFLW(Dataset):
    """ The WFLW dataset proposed by 

        Wu, Wayne, Chen Qian, Shuo Yang, Quan Wang, Yici Cai, and Qiang Zhou. 
        "Look at boundary: A boundary-aware face alignment algorithm." 
        In Proceedings of the IEEE conference on computer vision and 
        pattern recognition, pp. 2129-2138. 2018.

    It can be downloaded from https://wywu.github.io/projects/LAB/WFLW.html.

    """

    def __init__(self, root=None, split=Split.ALL, subset=None):
        if root is None:
            root = data_path.get('WFLW')

        self.root = root

        anno_file = None
        if split == Split.TRAIN:
            anno_file = 'list_98pt_rect_attr_train_test/list_98pt_rect_attr_train.txt'
        elif split == Split.TEST:
            if subset is None:
                anno_file = 'list_98pt_test/list_98pt_test.txt'
            else:
                anno_file = f'list_98pt_test/list_98pt_test_{subset}.txt'

        self.info_list = []
        with open(os.path.join(root, 'WFLW_annotations', anno_file), 'r') as f:
            for line_id, line in enumerate(f):
                line = line.strip()
                if len(line) == 0:
                    continue
                components = line.split()
                landmarks = np.reshape(
                    np.array([float(v) for v in components[:2*98]]), [98, 2])
                cx, cy = np.mean(landmarks, axis=0)
                image_file = components[-1]
                self.info_list.append({
                    'im_path': os.path.join(root, 'WFLW_images', image_file),
                    'sample_name': image_file.replace('/', '.') + ('_%.3f_%.3f' % (cx, cy)),
                    'landmarks': landmarks})

    def __len__(self):
        return len(self.info_list)

    def __getitem__(self, index):
        info = self.info_list[index]
        image = cv2.cvtColor(cv2.imread(info['im_path']), cv2.COLOR_BGR2RGB)
        x1, y1 = np.min(info['landmarks'], axis=0)
        x2, y2 = np.max(info['landmarks'], axis=0)
        box_y1x1y2x2 = np.array([y1, x1, y2, x2], dtype=np.float32)
        return {
            'image': image,
            'box': box_y1x1y2x2,
            'landmarks': info['landmarks']
        }

    def sample_name(self, index):
        return self.info_list[index]['sample_name']



import os
import numpy as np
import cv2
import re

import xmltodict

from . import data_path
from ..dataset import Dataset, Split


class IBUG300W(Dataset):
    """The IBUG-300W face alignment dataset.
    """

    def __init__(self, root=None, split=Split.ALL, subset='Fullset'):
        if root is None:
            root = data_path.get('300W', 'Splitted')

        self.root = root

        self.anno = []

        if split == Split.ALL:
            pattern = '.*'
        elif split == Split.TRAIN:
            pattern = '(afw/.*)|(helen/trainset/.*)|(lfpw/trainset/.*)'
        elif split == Split.TEST:
            pattern = {
                'Common': '(helen/testset/.*)|(lfpw/testset/.*)',
                'Challenging': 'ibug/.*',
                'Fullset': '(helen/testset/.*)|(lfpw/testset/.*)|(ibug/.*)',
            }[subset]
        else:
            raise RuntimeError(f'unsupported split {split} for IBUG300W')
        pattern = re.compile(pattern)

        with open(os.path.join(self.root, 'labels_ibug_300W.xml'), 'r') as fd:
            anno = xmltodict.parse(fd.read())

        self.info_list = []
        for info in anno['dataset']['images']['image']:
            if '_mirror' in info['@file']:
                continue
            if not pattern.match(info['@file']):
                continue

            im_path = os.path.join(self.root, info['@file'])

            sample_name = info['@file'].replace('/', '.')
            box_info = info['box']

            landmarks = np.zeros([68, 2], dtype=np.float32)
            for point in box_info['part']:
                index = int(point['@name'])
                landmarks[index, 0] = float(point['@x'])
                landmarks[index, 1] = float(point['@y'])
            self.info_list.append({
                'sample_name': sample_name,
                'im_path': im_path,
                'landmarks': landmarks
            })

    def __len__(self):
        return len(self.info_list)

    def __getitem__(self, index):
        info = self.info_list[index]
        image = cv2.cvtColor(cv2.imread(info['im_path']), cv2.COLOR_BGR2RGB)
        x1, y1 = np.min(info['landmarks'], axis=0)
        x2, y2 = np.max(info['landmarks'], axis=0)
        box_y1x1y2x2 = np.array([y1, x1, y2, x2], dtype=np.float32)
        return {
            'image': image,
            'box': box_y1x1y2x2,
            'landmarks': info['landmarks']
        }

    def sample_name(self, index):
        return self.info_list[index]['sample_name']



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import numpy as np
import cv2
import functools
from typing import Dict, List

from . import data_path
from ..dataset import Dataset, Split


@functools.lru_cache()
def _cached_imread(fname, flags=None):
    return cv2.imread(fname, flags=flags)


class CelebA(Dataset):
    """ The CelebA dataset proposed by 

        Liu, Ziwei, Ping Luo, Xiaogang Wang, and Xiaoou Tang. 
        "Deep learning face attributes in the wild." In Proceedings 
        of the IEEE international conference on computer vision, 
        pp. 3730-3738. 2015.

    It can be downloaded from https://mmlab.ie.cuhk.edu.hk/projects/CelebA.html.

    """

    def __init__(self, split=Split.ALL, root=None, aligned=False):
        if root is None:
            root = data_path.get('CelebA')
        if aligned:
            self.images_root = os.path.join(root, 'Img', 'img_align_celeba')
            self.anno_landmark_file = os.path.join(
                root, 'Anno', 'list_landmarks_align_celeba.txt')
        else:
            self.images_root = os.path.join(root, 'Img', 'img_celeba')
            self.anno_landmark_file = os.path.join(
                root, 'Anno', 'list_landmarks_celeba.txt')

        self.info_dict = dict()
        self.image_names = []
        with open(os.path.join(root, 'Eval', 'list_eval_partition.txt'), 'r') as f:
            for line in f:
                line = line.strip()
                if len(line) == 0:
                    continue
                image_name, eval_status = line.split()
                assert image_name.endswith('.jpg')
                eval_status = int(eval_status)
                if split == Split.TRAIN and eval_status == 0 or split == Split.VAL and eval_status == 1 \
                        or split == Split.TEST and eval_status == 2 or split == Split.ALL:
                    self.info_dict[image_name] = dict()
                    self.image_names.append(image_name)

        with open(os.path.join(root, 'Anno', 'identity_CelebA.txt'), 'r') as f:
            for line in f:
                line = line.strip()
                if len(line) == 0:
                    continue
                image_name, identity = line.split()
                if image_name in self.info_dict:
                    self.info_dict[image_name]['identity'] = int(identity)

        with open(os.path.join(root, 'Anno', 'list_attr_celeba.txt'), 'r') as f:
            for idx, line in enumerate(f):
                line = line.strip()
                if idx < 2 or len(line) == 0:
                    continue
                image_name, *attrs = line.split()
                if image_name in self.info_dict:
                    attrs = np.array(
                        [int(a) == 1 for a in attrs], dtype=np.uint8)
                    self.info_dict[image_name]['attrs'] = attrs

        if not aligned:
            with open(os.path.join(root, 'Anno', 'list_bbox_celeba.txt'), 'r') as f:
                for idx, line in enumerate(f):
                    line = line.strip()
                    if idx < 2 or len(line) == 0:
                        continue
                    image_name, *xywh = line.split()
                    if image_name in self.info_dict:
                        x, y, w, h = [int(v) for v in xywh]
                        y1x1y2x2 = np.array([y, x, y+h, x+w], dtype=np.float32)
                        self.info_dict[image_name]['box'] = y1x1y2x2

        with open(self.anno_landmark_file, 'r') as f:
            for idx, line in enumerate(f):
                line = line.strip()
                if idx < 2 or len(line) == 0:
                    continue
                image_name, *face_5_pts = line.split()
                if image_name in self.info_dict:
                    face_5_pts = np.reshape(
                        np.array([float(v) for v in face_5_pts], dtype=np.float32), [5, 2])
                    self.info_dict[image_name]['face_5_pts'] = face_5_pts

        # check all
        for image_name, info in self.info_dict.items():
            assert 'identity' in info
            assert 'attrs' in info
            if not aligned:
                assert 'box' in info
            assert 'face_5_pts' in info

    def __len__(self):
        return len(self.image_names)

    def __getitem__(self, index):
        image_name = self.image_names[index]
        image = cv2.cvtColor(_cached_imread(os.path.join(
            self.images_root, image_name)), cv2.COLOR_BGR2RGB)
        info = self.info_dict[image_name]
        data = {
            'image': image,
            'identity': info['identity'],
            'attrs': info['attrs'],
            'face_5_pts': info['face_5_pts']
        }
        if 'box' in info:
            data['box'] = info['box']
        return data

    def sample_name(self, index) -> str:
        return self.image_names[index][:-4]


class CelebAMaskHQ(Dataset):
    """ The CelebAMask-HQ dataset proposed by 

        Lee, Cheng-Han, Ziwei Liu, Lingyun Wu, and Ping Luo. 
        "Maskgan: Towards diverse and interactive facial image manipulation." 
        In Proceedings of the IEEE/CVF Conference on Computer Vision and 
        Pattern Recognition, pp. 5549-5558. 2020.

    It can be downloaded from https://github.com/switchablenorms/CelebAMask-HQ.
    """

    def __init__(self, split=Split.ALL, root=None, label_type='all'):
        if root is None:
            root = data_path.get('CelebAMask-HQ')
        assert os.path.exists(root)
        self.root = root
        self.split = split
        self.names = []

        if split != Split.ALL:
            hq_to_orig_mapping = dict()
            orig_to_hq_mapping = dict()
            mapping_file = os.path.join(
                root, 'CelebA-HQ-to-CelebA-mapping.txt')
            assert os.path.exists(mapping_file)
            for s in open(mapping_file, 'r'):
                if '.jpg' not in s:
                    continue
                idx, _, orig_file = s.split()
                hq_to_orig_mapping[int(idx)] = orig_file
                orig_to_hq_mapping[orig_file] = int(idx)

            # load partition
            partition_file = os.path.join(root, 'list_eval_partition.txt')
            assert os.path.exists(partition_file)
            for s in open(partition_file, 'r'):
                if '.jpg' not in s:
                    continue
                orig_file, group = s.split()
                group = int(group)
                if orig_file not in orig_to_hq_mapping:
                    continue
                hq_id = orig_to_hq_mapping[orig_file]
                if split == Split.TRAIN and group == 0:
                    self.names.append(str(hq_id))
                elif split == Split.VAL and group == 1:
                    self.names.append(str(hq_id))
                elif split == Split.TEST and group == 2:
                    self.names.append(str(hq_id))
                elif split == Split.TOY:
                    self.names.append(str(hq_id))
                    if len(self.names) >= 10:
                        break
        else:
            self.names = [
                n[:-(len('.jpg'))]
                for n in os.listdir(os.path.join(self.root, 'CelebA-HQ-img'))
                if n.endswith('.jpg')
            ]

        self.label_setting = {
            'human': {
                'suffix': [
                    'neck', 'skin', 'cloth', 'l_ear', 'r_ear', 'l_brow', 'r_brow',
                    'l_eye', 'r_eye', 'nose', 'mouth', 'l_lip', 'u_lip', 'hair'
                ],
                'names': [
                    'bg', 'neck', 'face', 'cloth', 'rr', 'lr', 'rb', 'lb', 're',
                    'le', 'nose', 'imouth', 'llip', 'ulip', 'hair'
                ]
            },
            'aux': {
                'suffix': [
                    'eye_g', 'hat', 'ear_r', 'neck_l',
                ],
                'names': [
                    'normal', 'glass', 'hat', 'earr', 'neckl'
                ]
            },
            'all': {
                'suffix': [
                    'neck', 'skin', 'cloth', 'l_ear', 'r_ear', 'l_brow', 'r_brow',
                    'l_eye', 'r_eye', 'nose', 'mouth', 'l_lip', 'u_lip', 'hair',
                    'eye_g', 'hat', 'ear_r', 'neck_l',
                ],
                'names': [
                    'bg', 'neck', 'face', 'cloth', 'rr', 'lr', 'rb', 'lb', 're',
                    'le', 'nose', 'imouth', 'llip', 'ulip', 'hair',
                    'glass', 'hat', 'earr', 'neckl'
                ]
            }
        }[label_type]

    def make_label(self, index, ordered_label_suffix):
        label = np.zeros((512, 512), np.uint8)
        name = self.names[index]
        name_id = int(name)
        name5 = '%05d' % name_id
        p = os.path.join(self.root, 'CelebAMask-HQ-mask-anno',
                         str(name_id // 2000), name5)
        for i, label_suffix in enumerate(ordered_label_suffix):
            label_value = i + 1
            label_fname = os.path.join(p + '_' + label_suffix + '.png')
            if os.path.exists(label_fname):
                mask = _cached_imread(label_fname, cv2.IMREAD_GRAYSCALE)
                label = np.where(mask > 0,
                                 np.ones_like(label) * label_value, label)
        return label

    def __getitem__(self, index):
        name = self.names[index]
        image = cv2.resize(
            cv2.imread(os.path.join(self.root, 'CelebA-HQ-img',
                                    name + '.jpg'))[:, :, ::-1],
            (512, 512),
            interpolation=cv2.INTER_LINEAR)

        data = {'image': image}
        label = self.make_label(index, self.label_setting['suffix'])
        data[f'label'] = label

        return data

    def __len__(self):
        return len(self.names)

    def sample_name(self, index):
        return self.names[index]

    @property
    def label_names(self) -> List[str]:
        return self.label_setting['names']



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import torch
import torch.nn as nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler as PyTorchScheduler
from timm.scheduler.scheduler import Scheduler as TimmScheduler

import re
from typing import List, Dict, Tuple, Any, Mapping, Optional, Union
import warnings

from .logger import Logger

Scheduler = Union[PyTorchScheduler, TimmScheduler]


class OptimizerConfig:
    """Configuration for both an optimizer and its lr-scheduler.

    Args:
        optimizer_type (str): The type name of the optimizer. e.g `torch.optim.AdamW`.
        optimizer_args (dict): The arguments of the optimizer.
        lr_scheduler_type (str): The type name of the lr_scheduler. e.g `torch.optim.lr_scheduler.MultiStepLR`.
        lr_scheduler_call (str): One of 'stepwise' or 'epochwise' or 'both.
        lr_scheduler_args (dict): The arguments of the lr_scheduler.
        network_settings (List[Dict[str, Any]]): The settings of networks to apply to.
            e.g. [{'params': 'main'}, {'params':'main.backbone', 'lr': 1e-3}]

    Usage:
        opt_config = OptimizerConfig(
            optimizer_type = 'torch.optim.AdamW',
            optimizer_args = {
                'lr': 0.001
                'betas': [0.9, 0.999]
                'weight_decay': 0.0001
            },
            lr_scheduler_type = 'torch.optim.lr_scheduler.ExponentialLR',
            lr_scheduler_args = {'gamma': 0.98},
            lr_scheduler_call = 'epochwise',
            network_settings = [
                {
                    'params': 'main.backbone.visual',
                    'lr': 0.0001
                }
            ]
        )

        optimizer, lr_scheduler = opt_config.instantiate(networks)
    """

    def __init__(self, networks: Mapping[str, nn.Module], optimizer_type: str, optimizer_args: dict,
                 lr_scheduler_type: str, lr_scheduler_args: dict,
                 network_settings: List[Dict[str, Any]],
                 lr_scheduler_call: str = 'epochwise',
                 steps_per_epoch: Optional[int] = None, logger: Optional[Logger] = None) -> None:

        # import haya
        self.optimizer_cfg = (eval(optimizer_type), optimizer_args)
        self.lr_scheduler_cfg = (eval(lr_scheduler_type), lr_scheduler_args)
        self.lr_scheduler_call = lr_scheduler_call
        assert lr_scheduler_call in {'stepwise', 'epochwise', 'both'}
        self.network_settings = network_settings
        assert isinstance(self.network_settings, list)

        self.optimizer, self.lr_scheduler = self._instantiate(
            networks, steps_per_epoch, logger)

    def scheduler_step(self, global_step):
        if self.lr_scheduler_call in {'stepwise', 'both'}:
            if isinstance(self.lr_scheduler, TimmScheduler):
                self.lr_scheduler.step_update(global_step+1)
            else:
                self.lr_scheduler.step()

    def scheduler_epoch(self, epoch):
        if self.lr_scheduler_call in {'epochwise', 'both'}:
            if isinstance(self.lr_scheduler, TimmScheduler):
                self.lr_scheduler.step(epoch)
            else:
                self.lr_scheduler.step()

    def load_state_dict(self, state):
        opt_state, sch_state = state
        self.optimizer.load_state_dict(opt_state)
        self.lr_scheduler.load_state_dict(sch_state)

    def state_dict(self):
        return (self.optimizer.state_dict(), self.lr_scheduler.state_dict())

    def _instantiate(self, networks: Mapping[str, nn.Module],
                     steps_per_epoch: Optional[int] = None,
                     logger: Optional[Logger] = None) -> Tuple[Optimizer, Scheduler]:
        """Instantiate the optimizer and the lr_scheduler for the networks.

        Args:
            networks (Mapping[str, nn.Module]): The networks to attach to. 
                e.g. {'main': SomeNetwork}
        """

        optimizer_type, optimizer_args = self.optimizer_cfg

        def _named_parameters(pattern: str) -> Tuple[List[torch.Tensor], List[str]]:
            collected_names = []
            collected_params = []
            for network_name, model in networks.items():
                if hasattr(model, '_ema_cfg'):
                    continue
                for param_name, param in model.named_parameters():
                    name = f'{network_name}.{param_name}'
                    if re.match(pattern, name):
                        collected_names.append(name)
                        collected_params.append(param)
            return collected_params, collected_names

        # interpret parameter settings
        param_groups = []

        all_collected_names = set()
        for network_setting in self.network_settings:
            assert isinstance(network_setting, dict)
            param_group = dict()
            for key, val in network_setting.items():
                if key == 'params':
                    assert isinstance(val, str)
                    val, names = _named_parameters(val)
                    if len(set(names) & all_collected_names) > 0:
                        raise RuntimeError(
                            f'Duplicated parameter settings for '
                            f'parameter names: {names & all_collected_names}')
                    all_collected_names |= set(names)

                param_group[key] = val
            param_groups.append(param_group)
        # check whether we lost some params
        all_existing_names = set(_named_parameters('.*')[1])
        if len(all_existing_names-all_collected_names) > 0:
            msg = (f'Some parameters are not included in parameter settings: '
                   f'{all_existing_names-all_collected_names}')
            if logger is None:
                warnings.warn(msg)
            else:
                logger.log_info(msg)

        optimizer: Optimizer = optimizer_type(
            param_groups, **optimizer_args)
        lr_scheduler_type, lr_scheduler_args = self.lr_scheduler_cfg

        # parse lr_scheduler_args
        _lr_scheduler_args = dict()
        for name, val in lr_scheduler_args.items():
            if isinstance(val, str) and '{steps_per_epoch}' in val:
                _lr_scheduler_args[name] = eval(
                    val.format(steps_per_epoch=steps_per_epoch))
                print(
                    f'NOTE: lr_scheduler_args[{name}]: {val} '
                    f'-> {_lr_scheduler_args[name]}')
            else:
                _lr_scheduler_args[name] = val

        lr_scheduler: Scheduler = lr_scheduler_type(
            optimizer, **_lr_scheduler_args)
        return optimizer, lr_scheduler



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools
import numpy as np
from typing import Tuple
import cv2
import random

from .basic import AugBase


class AsType(AugBase):
    """ x -> x
    """

    def __init__(self, dtype: np.dtype):
        self.dtype = dtype

    def process(self, x):
        return x.astype(self.dtype)


class ImRead(AugBase):
    """ path -> image
    """

    def process(self, path):
        return cv2.imread(path)


class Normalize255(AugBase):
    """ x -> clamp(float32(x / 255), 0, 1)
    """

    def process(self, x):
        result = (x / 255.0).astype(np.float32)
        result[result < 0] = 0
        result[result > 1] = 1
        return result


class Denormalize255(AugBase):
    """ x -> uint8(x * 255)
    """

    def process(self, x):
        result = (x * 255.0).astype(np.uint8)
        return result


class NormalizeImageNet(AugBase):
    def __init__(self):
        self.mean = np.array([0.485, 0.456, 0.406])    # rgb
        self.std = np.array([0.229, 0.224, 0.225])

    def process(self, im1):
        return (im1 - self.mean) / self.std


class HWC2CHW(AugBase):
    """ HWC -> CHW """

    def process(self, hwc):
        assert hwc.ndim == 3
        return np.transpose(hwc, (2, 0, 1))


class CHW2HWC(AugBase):
    """ CHW -> HWC """

    def process(self, chw):
        assert chw.ndim == 3
        return np.transpose(chw, (1, 2, 0))


class BGR2RGB(AugBase):
    """ BGR image -> RGB image
    """

    def process(self, bgr):
        assert bgr.ndim == 3
        return cv2.cvtColor(bgr, cv2.COLOR_BGR2RGB)


class RGB2BGR(AugBase):
    """ RGB image -> BGR image
    """

    def process(self, rgb):
        assert rgb.ndim == 3
        return cv2.cvtColor(rgb, cv2.COLOR_RGB2BGR)


class RGB2Gray(AugBase):
    """ RGB image -> Gray image (still 3 channels)
    """

    def process(self, rgb):
        assert rgb.ndim == 3
        im = cv2.cvtColor(rgb, cv2.COLOR_RGB2GRAY)
        return np.tile(np.expand_dims(im, -1), [1, 1, 3])


class FullLike(AugBase):
    """ array -> array
    """

    def __init__(self, fill_value):
        self.fill_value = fill_value

    def process(self, arr):
        return np.full_like(arr, self.fill_value)


class ArgMax(AugBase):
    """ value -> max_ids
    """

    def __init__(self, axis):
        self.axis = axis

    def process(self, data):
        return np.argmax(data, self.axis)


class RandomGray(AugBase):
    """ image -> image
    """

    def __init__(self, gray_ratio=0.1):
        self.gray_ratio = gray_ratio

    def process(self, image):
        if np.random.uniform() < self.gray_ratio:
            assert image.ndim == 3 and image.shape[-1] == 3
            image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
            image = np.tile(np.expand_dims(image, -1), [1, 1, 3])
        return image


class RandomGamma(AugBase):
    """ image -> image
    """

    def process(self, im1):
        assert im1.ndim == 3
        im1[im1 < 0] = 0
        im1[im1 > 1] = 1
        gamma = np.random.choice([-1, 0, 1])
        if gamma == -1:
            im1 = np.sqrt(im1)
        elif gamma == 1:
            im1 = np.square(im1)
        return im1


class RandomBlur(AugBase):
    """ image -> image
    """

    def __init__(self, scale=0.01):
        self.scale = scale

    def process(self, im):
        blur_kratio = np.random.uniform(low=0.0, high=self.scale)
        blur_ksize = int(
            (im.shape[0] + im.shape[1]) / 2 * blur_kratio)
        if blur_ksize > 1:
            im = np.clip(im, 0, 1.0)
            im = cv2.blur(im, (blur_ksize, blur_ksize))
        return im


@functools.lru_cache()
def _to_cv_inter(interp):
    table = {
        'bilinear': cv2.INTER_LINEAR,
        'nearest': cv2.INTER_NEAREST}
    return table[interp]


class SwapLabels(AugBase):
    """ label, label_names -> label_swapped, label_names_swapped
    """

    def __init__(self, swap_label_names=None):
        if swap_label_names is None:
            swap_label_names = [['lb', 'rb'], ['le', 're'], ['lr', 'rr']]
        self.swap_label_names = swap_label_names
        assert isinstance(self.swap_label_names, (tuple, list))

    def process(self, data):
        assert isinstance(data, (tuple, list))
        label, label_names = data
        assert label.ndim == 2
        label_swapped = np.copy(label)
        index_swapped = list(range(len(label_swapped)))
        # swap
        for lname, rname in self.swap_label_names:
            if lname in label_names and rname in label_names:
                lid = label_names.index(lname)
                rid = label_names.index(rname)
                label_swapped[label == lid] = rid
                label_swapped[label == rid] = lid
                index_swapped[lid] = rid
                index_swapped[rid] = lid
        label_names_swapped = [label_names[i] for i in index_swapped]
        return label_swapped, label_names_swapped


class ConvertLabels(AugBase):
    """ label -> converted_label
    """

    def __init__(self, from_label_names, to_label_names, outlier_value=-1):
        self.from_label_names = from_label_names
        self.to_label_names = to_label_names
        self.outlier_value = outlier_value

    def process(self, label):
        if self.from_label_names == self.to_label_names:
            return label
        new_label = np.ones_like(label) * self.outlier_value
        for to_label_ind, label_name in enumerate(self.to_label_names):
            if label_name in self.from_label_names:
                from_label_ind = self.from_label_names.index(label_name)
                new_label[label == from_label_ind] = to_label_ind
        return new_label


class ConvertLabelValues(AugBase):
    """ label -> converted_label

    Example:
        aug.ConvertLabelValues([{0}, {1, 3, 4}, {}, {2, 5}])
    """

    def __init__(self, convert_label_values, outlier_value=-1):
        self.convert_label_values = convert_label_values
        self.outlier_value = outlier_value

    def process(self, label):
        new_label = np.ones_like(label) * self.outlier_value
        for to_ind, from_inds in enumerate(self.convert_label_values):
            for from_ind in from_inds:
                new_label[label == from_ind] = to_ind
        return new_label


class ConvertLabelsToBinaryMask(AugBase):
    """ label -> mask(bool)

    Example:
        aug.ConvertLabelsToBinaryMask(positive_label_values:list)
    """

    def __init__(self, positive_label_values: Tuple[int, list, tuple]):
        self.positive_label_values = positive_label_values

    def process(self, label):
        if isinstance(self.positive_label_values, int):
            return (label == self.positive_label_values).astype(np.uint8)
        return np.logical_or.accumulate(
            [label == v for v in self.positive_label_values], dtype=np.uint8)


class ReplaceBackground(AugBase):
    """ image, fg_alpha -> image
    """

    def __init__(self, bg_dataset, bg_tag='image'):
        self.bg_dataset = bg_dataset
        self.bg_tag = bg_tag

    def process(self, data):
        image, fg_alpha = data
        bg_ind = np.random.randint(0, len(self.bg_dataset))
        bg_im = self.bg_dataset[bg_ind][self.bg_tag]

        # random bg image
        # fit bg to image
        h, w = image.shape[:2]

        scale_factor = max(h / bg_im.shape[0], w / bg_im.shape[1])
        bg_im = cv2.resize(
            bg_im, fx=scale_factor, fy=scale_factor, interpolation=cv2.INTER_LINEAR)
        assert bg_im.shape[0] >= h and bg_im.shape[1] >= w
        cx = np.random.randint(0, bg_im.shape[1] - w)
        cy = np.random.randint(0, bg_im.shape[0] - h)
        bg_im = bg_im[cy:cy + h, cx:cx + w, :] \
            if len(bg_im.shape) == 3 else bg_im[cy:cy + h, cx:cx + w]
        image = image * fg_alpha + \
            bg_im.astype(float) * (1.0 - fg_alpha)
        return image


class RandomPureColors(AugBase):
    """ image_shape -> image: float, [0~1]
    """

    def __init__(self, low=0.0, high=1.0):
        self.low = low
        self.high = high

    def process(self, image_shape):
        h, w, c = image_shape
        color = np.random.uniform(self.low, self.high, size=[1, 1, c])
        return np.broadcast_to(color, [h, w, c])


class RandomPixels(AugBase):
    """ image_shape -> image: float, [0~1]
    """

    def __init__(self, low=0.0, high=1.0):
        self.low = low
        self.high = high

    def process(self, image_shape):
        return np.random.uniform(self.low, self.high, size=image_shape)


class RandomOcclusion(AugBase):
    """ image -> image
    """

    def __init__(self, h_min: float = 0.2, h_max: float = 0.8, w_delta: float = 0.2):
        self.h_min = h_min
        self.h_max = h_max
        self.w_delta = w_delta

    def process(self, image):
        h, w, _ = image.shape
        rh = random.random() * (self.h_max - self.h_min) + self.h_min
        rw = rh - self.w_delta + 2 * self.w_delta * random.random()
        cx = int((h - 1) * random.random())
        cy = int((w - 1) * random.random())
        dh = int(h / 2 * rh)
        dw = int(w / 2 * rw)
        x0 = max(0, cx - dw // 2)
        y0 = max(0, cy - dh // 2)
        x1 = min(w - 1, cx + dw // 2)
        y1 = min(h - 1, cy + dh // 2)
        image = np.array(image)
        image[y0:y1+1, x0:x1+1, :] = 0
        return image


class NoiseFusion(AugBase):
    """ image -> image
    """

    def process(self, image):
        h, w, c = image.shape
        dtype = image.dtype
        noise = np.random.rand(h, w, c)
        alpha = 0.5 * random.random()
        image = (1 - alpha) * image + alpha * noise
        return image.astype(dtype)


class FlipImage(AugBase):
    """ image -> image
    """

    def __init__(self, axis=1):
        self.axis = axis

    def process(self, image):
        return np.array(np.flip(image, axis=self.axis))


class DrawPointsOnImage(AugBase):
    """ image (hxwx3), points (nx2) -> image """

    def __init__(self, radius, color=None, with_text=False, font_scale=0.2):
        self.radius = radius
        if color is None:
            color = [255] * 4
        self.color = color
        self.with_text = with_text
        self.font_scale = font_scale

    def process(self, data):
        image, points = data
        image = np.array(image)
        points = points.astype(np.int32)
        for i, (x, y) in enumerate(points):
            cv2.circle(image, (x, y), self.radius, self.color, -1, cv2.LINE_AA)
            if self.with_text:
                cv2.putText(image, str(i), (x+1, y-1),
                            cv2.FONT_HERSHEY_SCRIPT_SIMPLEX,
                            self.font_scale, self.color, None, cv2.LINE_AA)
        return image



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import colorsys
import random
import numpy as np

import cv2
from matplotlib import pyplot as plt
from skimage import draw


def gen_random_colors(N, bright=True):
    brightness = 1.0 if bright else 0.7
    hsv = [(i / N, 1, brightness) for i in range(N)]
    colors = list(map(lambda c: colorsys.hsv_to_rgb(*c), hsv))
    random.shuffle(colors)
    return colors


_static_label_colors = [
    np.array((1.0, 1.0, 1.0), np.float32),
    np.array((255, 250, 79), np.float32) / 255.0,  # face
    np.array([255, 125, 138], np.float32) / 255.0,  # lb
    np.array([213, 32, 29], np.float32) / 255.0,  # rb
    np.array([0, 144, 187], np.float32) / 255.0,  # le
    np.array([0, 196, 253], np.float32) / 255.0,  # re
    np.array([255, 129, 54], np.float32) / 255.0,  # nose
    np.array([88, 233, 135], np.float32) / 255.0,  # ulip
    np.array([0, 117, 27], np.float32) / 255.0,  # llip
    np.array([255, 76, 249], np.float32) / 255.0,  # imouth
    np.array((1.0, 0.0, 0.0), np.float32),
    np.array((255, 250, 79), np.float32) / 255.0,
    np.array((255, 250, 79), np.float32) / 255.0,
    np.array((250, 245, 74), np.float32) / 255.0,
    np.array((0.0, 1.0, 0.5), np.float32),
    np.array((1.0, 0.0, 0.5), np.float32),
] + gen_random_colors(256)

_names_in_static_label_colors = [
    'bg', 'face', 'lb', 'rb', 'le', 're', 'nose',
    'ulip', 'llip', 'imouth', 'hair', 'lr', 'rr', 'neck',
    'cloth', 'eyeg', 'hat', 'earr'
]


def get_static_colors(size=0):
    if size <= len(_static_label_colors):
        return _static_label_colors
    return _static_label_colors + gen_random_colors(
        size-len(_static_label_colors))


def blend_labels(image, labels, label_names_dict=None,
                 default_alpha=0.6, color_offset=None):
    assert labels.ndim == 2
    bg_mask = labels == 0
    if label_names_dict is None:
        colors = _static_label_colors
    else:
        colors = [np.array((1.0, 1.0, 1.0), np.float32)]
        for i in range(1, labels.max() + 1):
            if isinstance(label_names_dict, dict) and i not in label_names_dict:
                bg_mask = np.logical_or(bg_mask, labels == i)
                colors.append(np.zeros((3)))
                continue
            label_name = label_names_dict[i]
            if label_name in _names_in_static_label_colors:
                color = _static_label_colors[
                    _names_in_static_label_colors.index(
                        label_name)]
            else:
                color = np.array((1.0, 1.0, 1.0), np.float32)
            colors.append(color)

    if color_offset is not None:
        ncolors = []
        for c in colors:
            nc = np.array(c)
            if (nc != np.zeros(3)).any():
                nc += color_offset
            ncolors.append(nc)
        colors = ncolors

    if image is None:
        image = orig_image = np.zeros(
            [labels.shape[0], labels.shape[1], 3], np.float32)
        alpha = 1.0
    else:
        orig_image = image / np.max(image)
        image = orig_image * (1.0 - default_alpha)
        alpha = default_alpha
    for i in range(1, np.max(labels) + 1):
        image += alpha * \
            np.tile(
                np.expand_dims(
                    (labels == i).astype(np.float32), -1),
                [1, 1, 3]) * colors[(i) % len(colors)]
    image[np.where(image > 1.0)] = 1.0
    image[np.where(image < 0)] = 0.0
    image[np.where(bg_mask)] = orig_image[np.where(bg_mask)]
    return image


_default_color = np.array([1, 0, 0], np.float32)


def draw_line(image, y1, x1, y2, x2, thickness, color=_default_color):
    if image.dtype == np.uint8:
        image = image / 255.0
    else:
        image = np.copy(image)

    perp = np.array([y1 - y2, x2 - x1])
    perp = perp * thickness / 2.0 / np.linalg.norm(perp)

    p1 = np.array([x1, y1])
    p2 = np.array([x2, y2])

    p1a = p1 - perp
    p1b = p1 + perp
    p2a = p2 - perp
    p2b = p2 + perp

    r = np.array([p1a[1], p2a[1], p2b[1], p1b[1]])
    c = np.array([p1a[0], p2a[0], p2b[0], p1b[0]])
    image[draw.polygon(r, c, shape=image.shape)] = color
    return image


def draw_boxes_y1x1y2x2(image, boxes, thickness, color=_default_color):
    r"""Draw boxes on image

    Args:
        image: A numpy array with shape h x w x 3, dtype uint8(0-255) or float(0-1)
        boxes: A numpy array with shape 4 or num_boxes x 4, dtype int

    Returns:
        image: h x w x 3, float(0-1)
    """
    if image.dtype == np.uint8:
        image = image / 255.0
    else:
        image = np.copy(image)
    if boxes.ndim == 1:
        boxes = np.expand_dims(boxes, 0)
    for i in range(boxes.shape[0]):
        y1, x1, y2, x2 = boxes[i]
        y1 = int(y1)
        x1 = int(x1)
        y2 = int(y2)
        x2 = int(x2)
        image = draw_line(image, y1 - thickness / 2, x1 - thickness / 2,
                          y1 - thickness / 2, x2 + thickness / 2,
                          thickness, color)
        image = draw_line(image, y1 - thickness / 2, x1 - thickness / 2,
                          y2 + thickness / 2, x1 - thickness / 2,
                          thickness, color)
        image = draw_line(image, y2 + thickness / 2, x1 - thickness / 2,
                          y2 + thickness / 2, x2 + thickness / 2,
                          thickness, color)
        image = draw_line(image, y1 - thickness / 2, x2 + thickness / 2,
                          y2 + thickness / 2, x2 + thickness / 2,
                          thickness, color)
    return image


def draw_circle(image, center_y, center_x, radius, color=_default_color):
    if image.dtype == np.uint8:
        image = image / 255.0
    else:
        image = np.copy(image)
    image[draw.circle(center_y, center_x, radius, shape=image.shape)] = color
    return image


def draw_points_yx(image, points, radius, color=_default_color):
    if image.dtype == np.uint8:
        image = image / 255.0
    else:
        image = np.copy(image)
    if points.ndim == 1:
        points = np.expand_dims(points, 0)
    for i in range(points.shape[0]):
        cy, cx = points[i]
        image = draw_circle(image, cy, cx, radius, color)
    return image


def draw_points_xy(image, points, radius, color=_default_color):
    if image.dtype == np.uint8:
        image = image / 255.0
    else:
        image = np.copy(image)
    if points.ndim == 1:
        points = np.expand_dims(points, 0)
    for i in range(points.shape[0]):
        cx, cy = points[i]
        image = draw_circle(image, cy, cx, radius, color)
    return image


def draw_landmarks(image, landmarks, color=None, radius=3,
                   with_text=False, font_scale=1.0, font_color=None):
    landmarks = landmarks.astype(np.int32)
    if color is None:
        color = [255] * 4
    if font_color is None:
        font_color = [255]
    for i, (x, y) in enumerate(landmarks):
        if callable(color):
            c = color(i)
        else:
            c = color
        cv2.circle(image, (x, y), radius, c, -1)
        if with_text:
            cv2.putText(image, str(i), (x+1, y-1),
                        cv2.FONT_HERSHEY_SCRIPT_SIMPLEX, font_scale, font_color)
    return image


def set_plt_figsize(a, b):
    plt.rcParams['figure.figsize'] = (a, b)


def show_single(image, **kwargs):
    print(f'showing images with shape={image.shape}, '
          f'dtype={image.dtype}, max={image.max()}, min={image.min()}')
    plt.imshow(image, **kwargs)
    # plt.pause(0.1)
    plt.show()


def show_batch(images, nrows=1, **kwargs):
    assert images.ndim in [3, 4]
    if images.ndim == 3:
        images = np.expand_dims(images, -1)
    print(f'showing images with shape={images.shape}, '
          f'dtype={images.dtype}, max={images.max()}, min={images.min()}')
    for i, image in enumerate(images):
        plt.subplot(nrows, (len(images)+nrows-1)//nrows, i+1, frameon=True)
        plt.imshow(image, **kwargs)
    plt.show()


def show_multi(*images, nrows=1, **kwargs):
    if len(images) == 0:
        print('input is empty')
        return
    for i, image in enumerate(images):
        print(f'showing image[{i}] with shape={image.shape}, '
              f'dtype={image.dtype}, max={image.max()}, min={image.min()}')
        plt.subplot(nrows, (len(images)+nrows-1)//nrows, i+1, frameon=True)
        plt.imshow(image, **kwargs)
    plt.show()




# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Mapping, Any, Tuple, List
from dataclasses import dataclass

import torch
from torch import nn
import torch.nn.functional as F


@dataclass
class ForwardFlags:
    with_losses: bool = True  # for training
    with_outputs: bool = True  # for measurement
    with_images: bool = True  # for visualization


class Task(nn.Module):
    """A task module that computes losses and outputs with given networks and data.
    """

    def setup_networks(self, networks: Mapping[str, nn.Module]):
        """Setup networks.

        Attach those networks whose parameters will be updated.
        """
        pass

    def forward_networks(self, data: Mapping[str, torch.Tensor]
                         ) -> Mapping[str, torch.Tensor]:
        raise NotImplementedError()

    def compute_losses_and_outputs(
            self, data: Mapping[str, torch.Tensor],
            flags: ForwardFlags) -> Tuple[torch.Tensor,
                                          Mapping[str, torch.Tensor],
                                          Mapping[str, torch.Tensor],
                                          Mapping[str, torch.Tensor]]:
        raise NotImplementedError()

    def forward(self, data: Mapping[str, torch.Tensor], flags: ForwardFlags
                ) -> Tuple[torch.Tensor,
                           Mapping[str, torch.Tensor],
                           Mapping[str, torch.Tensor],
                           Mapping[str, torch.Tensor]]:
        """Forwarding with one batch of data, and returns corresponding losses/scores and outputs.

        Args:
            data (Mapping[str, torch.Tensor]): One data batch.
            with_outputs (bool): Whether to compute the outputs for visualization.
            flags (ForwardFlags): Flags for computation.

        Returns:
            Tuple[torch.Tensor, Mapping[str, torch.Tensor], Mapping[str, torch.Tensor], Mapping[str, torch.Tensor]]:
                The total loss/score, the named losses/scores, outputs (can be None) and output_images.
        """
        raise NotImplementedError()



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import torch
import torch.utils
import torch.utils.data
import bisect
from enum import Enum


class Split(Enum):
    ALL = 0
    TRAIN = 1
    VAL = 2
    TEST = 3
    REFINED_TEST = 4
    TOY = 5


ALL = Split.ALL
TRAIN = Split.TRAIN
VAL = Split.VAL
TEST = Split.TEST
REFINED_TEST = Split.REFINED_TEST
TOY = Split.TOY


class Dataset(torch.utils.data.Dataset):
    """The augmentable dataset. All augmentable dataset should derive this class.

    Comparing to torch.utils.data.Dataset, this class should further implement the `sample_name` method.
    """

    def augment(self, aug):
        from .augmenters import augment
        return augment(self, aug)

    def sample_name(self, index) -> str:
        """The name of this sample.
        """
        raise NotImplementedError()

    def extra_data(self) -> dict:
        return dict()

    def subset(self, indices):
        return Subset(self, indices)

    def attach_index_as_tag(self):
        return _DatasetWithIndexAsTag(self)

    def attach_sample_name_as_tag(self):
        return _DatasetWithSampleNameAsTag(self)

    def validate_sample_names(self, valid_sample_names):
        return _DatasetWithValidSampleNames(self, valid_sample_names)

    def invalidate_sample_names(self, invalid_sample_names):
        return _DatasetWithInvalidSampleNames(self, invalid_sample_names)

    def make_evenly_divisible(self, divisor: int, validity_tag: str = 'valid'):
        return _DatasetEvenlyDivisible(self, divisor, validity_tag)

    def collect_as_dict(self, tag_to_collect, with_prog=False):
        collected = dict()
        if with_prog:
            from tqdm.auto import tqdm
            p = tqdm(self)
        else:
            p = self
        for i, d in enumerate(p):
            collected[self.sample_name(i)] = d[tag_to_collect]
        return collected


class Subset(Dataset):
    """Subset of a dataset at specified indices.

    Args:
        dataset (Dataset): The whole Dataset
        indices (list): Indices in the whole set selected for subset
    """

    def __init__(self, dataset, indices):
        self.dataset = dataset
        self.indices = indices

    def __getitem__(self, idx):
        return self.dataset[self.indices[idx]]

    def sample_name(self, idx):
        return self.dataset.sample_name(self.indices[idx])

    def __len__(self):
        return len(self.indices)

    def extra_data(self) -> dict:
        return self.dataset.extra_data()


class ConcatDataset(Dataset):
    """Dataset as a concatenation of multiple datasets.

    This class is useful to assemble different existing datasets.

    Args:
        datasets: List of datasets to be concatenated
    """

    @staticmethod
    def cumsum(sequence):
        r, s = [], 0
        for e in sequence:
            l = len(e)
            r.append(l + s)
            s += l
        return r

    def __init__(self, datasets):
        assert len(datasets) > 0, 'datasets should not be an empty iterable'
        self.datasets = list(datasets)
        self.cumulative_sizes = self.cumsum(self.datasets)

    def __len__(self):
        return self.cumulative_sizes[-1]

    def _get_dataset_sample_inds(self, idx):
        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    "absolute value of index should not exceed dataset length")
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return dataset_idx, sample_idx

    def __getitem__(self, idx):
        dataset_idx, sample_idx = self._get_dataset_sample_inds(idx)
        return self.datasets[dataset_idx][sample_idx]

    def sample_name(self, idx):
        dataset_idx, sample_idx = self._get_dataset_sample_inds(idx)
        return self.datasets[dataset_idx].sample_name(sample_idx)

    def extra_data(self) -> dict:
        result = dict()
        for ds in self.datasets:
            result.update(ds.extra_data())
        return result


class _DatasetWithIndexAsTag(Dataset):
    def __init__(self, orig_dataset, tag='index'):
        self.orig_dataset = orig_dataset
        self.tag = tag

    def sample_name(self, index):
        return self.orig_dataset.sample_name(index)

    def __len__(self):
        return len(self.orig_dataset)

    def __getitem__(self, index):
        return {self.tag: index, **self.orig_dataset[index]}

    def extra_data(self) -> dict:
        return self.orig_dataset.extra_data()


class _DatasetWithSampleNameAsTag(Dataset):
    def __init__(self, orig_dataset, tag='sample_name'):
        self.orig_dataset = orig_dataset
        self.tag = tag

    def sample_name(self, index):
        return self.orig_dataset.sample_name(index)

    def __len__(self):
        return len(self.orig_dataset)

    def __getitem__(self, index):
        return {self.tag: self.orig_dataset.sample_name(index),
                **self.orig_dataset[index]}

    def extra_data(self) -> dict:
        return self.orig_dataset.extra_data()


class _DatasetWithValidSampleNames(Dataset):
    def __init__(self, orig_dataset, valid_sample_names):
        self.orig_dataset = orig_dataset
        self.valid_inds = []
        for i in range(len(orig_dataset)):
            if orig_dataset.sample_name(i) in valid_sample_names:
                self.valid_inds.append(i)

    def sample_name(self, index):
        return self.orig_dataset.sample_name(self.valid_inds[index])

    def __len__(self):
        return len(self.valid_inds)

    def __getitem__(self, index):
        return self.orig_dataset[self.valid_inds[index]]

    def extra_data(self) -> dict:
        return self.orig_dataset.extra_data()


class _DatasetWithInvalidSampleNames(Dataset):
    def __init__(self, orig_dataset, invalid_sample_names):
        self.orig_dataset = orig_dataset
        self.valid_inds = []
        for i in range(len(orig_dataset)):
            if orig_dataset.sample_name(i) not in invalid_sample_names:
                self.valid_inds.append(i)

    def sample_name(self, index):
        return self.orig_dataset.sample_name(self.valid_inds[index])

    def __len__(self):
        return len(self.valid_inds)

    def __getitem__(self, index):
        return self.orig_dataset[self.valid_inds[index]]

    def extra_data(self) -> dict:
        return self.orig_dataset.extra_data()


class _DatasetEvenlyDivisible(Dataset):
    def __init__(self, dataset: Dataset, divisor: int, validity_tag: str = 'valid'):
        self.dataset = dataset
        self.divisible_len = len(self.dataset)
        if self.divisible_len % divisor > 0:
            self.divisible_len += (divisor-self.divisible_len % divisor)
        assert self.divisible_len % divisor == 0
        assert self.divisible_len >= len(self.dataset)
        self.validity_tag = validity_tag

    def __len__(self):
        return self.divisible_len

    def __getitem__(self, index):
        if index < 0:
            index += self.divisible_len
        data = self.dataset[index % len(self.dataset)]
        return {self.validity_tag: (index < len(self.dataset)), **data}

    def sample_name(self, index) -> str:
        if index < 0:
            index += self.divisible_len
        return self.dataset.sample_name(index % len(self.dataset))

    def extra_data(self) -> dict:
        return self.dataset.extra_data()


class DatasetRepeatingConstantValue(Dataset):
    def __init__(self, value, length: int):
        self.value = value
        self.length = length

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        return self.value

    def sample_name(self, index) -> str:
        return f'constant_{index}'



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import numpy as np
import cv2
import scipy.io

from . import data_path
from ..dataset import Dataset, Split


class AFLW_19(Dataset):
    """ The AFLW-19 dataset used by [1], which has 19 landmark points in each image. 
    Note that it is NOT the original AFLW [2] dataset which has 21 landmark points per image.

    AFLW-19 data can be downloaded from http://mmlab.ie.cuhk.edu.hk/projects/compositional.html.

    Citations:

        [1] Zhu, Shizhan, Cheng Li, Chen-Change Loy, and Xiaoou Tang. 
            "Unconstrained face alignment via cascaded compositional learning." 
            In Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition, 
            pp. 3409-3417. 2016.
        
        [2] Koestinger, Martin, Paul Wohlhart, Peter M. Roth, and Horst Bischof. 
            "Annotated facial landmarks in the wild: A large-scale, real-world database for facial 
            landmark localization." In 2011 IEEE international conference on computer vision workshops 
            (ICCV workshops), pp. 2144-2151. IEEE, 2011.
    """

    def __init__(self, root=None, split=Split.ALL, subset: str = 'full'):
        if root is None:
            root = data_path.get('AFLW')
        self.images_root = os.path.join(root, 'aflw-21', 'data', 'flickr')
        info = scipy.io.loadmat(os.path.join(
            root, 'aflw-19', 'AFLWinfo_release.mat'))
        self.bbox = info['bbox']  # 24386x4 left, right, top bottom
        self.data = info['data']  # 24386x38 x1,x2...,xn,y1,y2...,yn
        self.mask = info['mask_new']  # 24386x19
        self.name_list = [s[0][0] for s in info['nameList']]

        ra = np.reshape(info['ra'].astype(np.int32), [-1])-1
        assert ra.min() == 0
        assert ra.max() == self.bbox.shape[0] - 1
        if split == Split.ALL:
            self.indices = ra
        elif split == Split.TRAIN:
            self.indices = ra[:20000]
        elif split == Split.TEST:
            if subset == 'full':
                self.indices = ra[20000:]
            elif subset == 'frontal':
                all_visible = np.all(self.mask == 1, axis=1)  # 24386
                self.indices = np.array(
                    [ind for ind in ra[20000:] if all_visible[ind]])

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, index):
        ind = self.indices[index]
        image = cv2.cvtColor(cv2.imread(os.path.join(
            self.images_root, self.name_list[ind])), cv2.COLOR_BGR2RGB)
        landmarks = np.reshape(self.data[ind], [2, 19]).transpose()

        left, right, top, bottom = self.bbox[ind]
        box_y1x1y2x2 = np.array([top, left, bottom, right], dtype=np.float32)

        visibility = self.mask[ind]
        return {
            'image': image,
            'box': box_y1x1y2x2,
            'landmarks': landmarks,
            'visibility': visibility
        }

    def sample_name(self, index):
        return str(index)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from typing import Mapping, List, Dict, Optional, Any, Union, Set
from contextlib import contextmanager

import torch
from torch import nn
import torch.distributed as dist

from ..context import Context

from .ema_config import EMAConfig, _interpret_ema_networks

from .dataprocessor import DataProcessor
from .logger import Logger
from .task import ForwardFlags, Task
from .scorer import Scorer
from .optimizer import OptimizerConfig
from .stats import StatSummary
from .util import _load_states, _save_states, barrier, accumulate


class Trainer:
    """A standard trainer for supervised training. You should use DistributedGPURun with it.
    """

    def __init__(self,
                 states_dir: str,
                 outputs_dir: str,
                 train_dataprocessor: DataProcessor,
                 eval_dataprocessors: Mapping[str, DataProcessor],
                 networks: Mapping[str, Union[nn.Module, EMAConfig]],
                 task_module: Task,
                 scorer: Scorer,
                 optimizer_cfg: dict,
                 logger: Logger,
                 cont_opti: bool = True,
                 num_states_kept: Optional[int] = None,
                 max_epoches: Optional[int] = None,
                 enable_amp: bool = False,
                 states_save_interval: int = 1,
                 eval_interval: int = 1,
                 states_save_extra_steps: List[int] = [],
                 _ctx: Optional[Context] = None):

        self.context = _ctx

        self.outputs_dir = outputs_dir
        self.states_dir = states_dir

        self.train_dataprocessor = train_dataprocessor
        self.eval_dataprocessors = eval_dataprocessors

        self.networks = _interpret_ema_networks(networks)

        self.task_module = task_module
        self.scorer = scorer

        self.logger = logger
        self.logger.setup(self.outputs_dir)
        self.logger.log_info(f'outputs_dir: {self.outputs_dir}')
        self.logger.log_info(f'states_dir: {self.states_dir}')

        self.optimizer_cfg = OptimizerConfig(
            networks=self.networks,
            steps_per_epoch=self.train_dataprocessor.num_steps_of_epoch(0),
            logger=logger, **optimizer_cfg)

        self.rank = self.world_size = None

        self.cont_opti = cont_opti
        self.num_states_kept = num_states_kept

        self.max_epoches = max_epoches
        self.states_save_interval = states_save_interval
        self.eval_interval = eval_interval
        self.states_save_extra_steps = states_save_extra_steps

        self.enable_amp = enable_amp
        self.amp_scaler = None
        if self.enable_amp:
            self.amp_scaler = torch.cuda.amp.GradScaler()

        self.global_step = None
        self.epoch = None

        self.world_size = self.rank = None
        self.debug_loss = _ctx.debug_loss
        self.detect_anomaly = _ctx.detect_anomaly

    @contextmanager
    def _train_scope(self):
        with torch.cuda.amp.autocast(enabled=self.enable_amp):
            with torch.autograd.set_detect_anomaly(self.detect_anomaly):
                yield

    def _backward(self, loss: torch.Tensor):
        if self.amp_scaler is None:
            loss.backward()
        else:
            self.amp_scaler.scale(loss).backward()

    def _optimizer_step(self):
        if self.amp_scaler is None:
            self.optimizer_cfg.optimizer.step()
        else:
            self.amp_scaler.step(self.optimizer_cfg.optimizer)
            self.amp_scaler.update()

        # update ema modules
        for _, net in self.networks.items():
            if hasattr(net, '_ema_cfg'):
                ema_cfg: EMAConfig = net._ema_cfg
                orig_net = self.networks[ema_cfg.network_name]
                decay = ema_cfg.decay
                accumulate(net, orig_net, decay=decay)

    def __call__(self):
        if not dist.is_initialized():
            raise RuntimeError(
                "Requires torch.distributed package to be initialized")
        self.rank = dist.get_rank()
        assert self.rank is not None
        self.world_size = dist.get_world_size()
        assert self.world_size is not None

        if self.rank == 0:
            os.makedirs(self.outputs_dir, exist_ok=True)
            os.makedirs(self.states_dir, exist_ok=True)

        self.epoch = 0
        self.global_step = 0

        self._load_train_states()

        self.task_module.setup_networks(self.networks)
        task_module = nn.parallel.DistributedDataParallel(
            self.task_module.float().cuda(), device_ids=[torch.cuda.current_device()],
            find_unused_parameters=True)

        while self.max_epoches is None or self.epoch < self.max_epoches:
            barrier()

            task_module.train()
            loss_summary_this_epoch = StatSummary()

            self.optimizer_cfg.optimizer.zero_grad()
            num_steps = self.train_dataprocessor.num_steps_of_epoch(self.epoch)
            self.logger.log_info(
                f'There will be {num_steps} training steps in this epoch.')

            for step, data in enumerate(self.train_dataprocessor.iterate_input(self.epoch)):
                time_to_save_outputs = step == num_steps - 1 or self.global_step == 0
                if step >= num_steps:
                    break

                if self.global_step in self.states_save_extra_steps:
                    self._save_train_states(losses=None)

                # forward
                with self._train_scope():
                    loss, named_losses, _, images = task_module(
                        data, ForwardFlags(with_losses=True,
                                           with_outputs=False,
                                           with_images=time_to_save_outputs))

                    if self.debug_loss:
                        print(f'loss={loss.item()}')

                # backward
                self.optimizer_cfg.optimizer.zero_grad()
                # self.amp_scaler.
                self._backward(loss)

                # update
                self._optimizer_step()
                self.optimizer_cfg.scheduler_step(self.global_step)

                # accumulate loss values
                loss_summary_this_epoch.update(named_losses)

                # save output images at the end of each training epoch
                if time_to_save_outputs:
                    self.logger.log_numbers(
                        loss_summary_this_epoch.get_averages(),
                        global_step=self.global_step, epoch=self.epoch, tag='train')
                    self.logger.log_batched_images(
                        images, global_step=self.global_step, epoch=self.epoch, tag='train')

                self.global_step += 1

            self.optimizer_cfg.scheduler_epoch(self.epoch)

            self.epoch += 1

            barrier()

            if self.epoch % self.states_save_interval == 0:
                self._save_train_states(losses=None)

            if self.epoch % self.eval_interval == 0:
                task_module.eval()
                # evaluate on the eval_dataprocessor using the scorer
                for eval_name, eval_dataprocessor in self.eval_dataprocessors.items():
                    self.logger.log_info(
                        f'evaluating {eval_dataprocessor.num_steps_of_epoch(0)} '
                        f'batches from {eval_name} ...')
                    self.scorer.init_evaluation()
                    for result_data in eval_dataprocessor.iterate_final_result(self.task_module):
                        self.scorer.evaluate(result_data)
                    scores = self.scorer.finalize_evaluation()
                    self.logger.log_numbers(scores,
                                            global_step=self.global_step,
                                            epoch=self.epoch, tag=f'eval.{eval_name}')

                # show GPU status
                if self.epoch == self.eval_interval and torch.cuda.current_device() == 0:
                    os.system('nvidia-smi')

    def _load_train_states(self):
        states = _load_states(self.states_dir, self.logger)
        if states is not None:
            self.epoch = states['epoch']
            self.global_step = states['global_step']
            for net_name, net in self.networks.items():
                net.load_state_dict(states['networks'][net_name])
            if self.cont_opti:
                self.optimizer_cfg.load_state_dict(states['optimizer'])
            del states

    def _save_train_states(self, losses=None):
        if self.rank == 0:
            if losses is None:
                losses = dict()
            states = {'epoch': self.epoch, 'global_step': self.global_step,
                      'losses': losses}
            states['networks'] = {name: net.state_dict()
                                  for name, net in self.networks.items()}
            states['optimizer'] = self.optimizer_cfg.state_dict()

            _save_states(states, self.states_dir,
                         losses=losses,
                         global_step=self.global_step, epoch=self.epoch,
                         num_states_kept=self.num_states_kept,
                         always_kept_steps=[],
                         always_kept_epoches=[],
                         logger=self.logger)
            self.logger.log_info(
                'all states saved with global_step=%d' % self.global_step)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Mapping, List

import numpy as np


class Scorer:
    """A base class for scoring a collection of data.
    """

    def init_evaluation(self):
        """Initialize evaluation.
        """
        pass

    def evaluate(self, data: Mapping[str, np.ndarray]):
        """Evaluate on data sample.

        Args:
            data (Mapping[str, np.ndarray]): One data sample.
        """
        raise NotImplementedError()

    def finalize_evaluation(self) -> Mapping[str, float]:
        """Finalize the evaluation.

        Returns:
            Mapping[str, float]: The final results.
        """
        raise NotImplementedError()


class MultipleScorers(Scorer):
    def __init__(self, scorers: List[Scorer]) -> None:
        self.scorers = scorers

    def init_evaluation(self):
        for s in self.scorers:
            s.init_evaluation()

    def evaluate(self, data: Mapping[str, np.ndarray]):
        for s in self.scorers:
            s.evaluate(data)

    def finalize_evaluation(self) -> Mapping[str, float]:
        merged_outputs = dict()
        for s in self.scorers:
            merged_outputs.update(s.finalize_evaluation())
        return merged_outputs



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import numpy as np
import cv2

from . import data_path
from ..dataset import Dataset, Split


class LaPa(Dataset):
    """The LaPa face parsing dataset proposed in 

        Liu, Yinglu, Hailin Shi, Hao Shen, Yue Si, Xiaobo Wang, and Tao Mei. 
        "A new dataset and boundary-attention semantic segmentation for face parsing." 
        In Proceedings of the AAAI Conference on Artificial Intelligence, 
        vol. 34, no. 07, pp. 11637-11644. 2020.

    It can be downloaded from https://github.com/JDAI-CV/lapa-dataset/.

    Args:
        root (str): The directory that contains subdirs 'image', 'labels'.
    """

    def __init__(self, root=None, split=Split.ALL):
        if root is None:
            root = data_path.get('LaPa')
        assert root is not None
        self.root = root

        subfolders = []
        if split == Split.TRAIN:
            subfolders = ['train']
        elif split == Split.VAL:
            subfolders = ['val']
        elif split in {Split.TEST, Split.TOY}:
            subfolders = ['test']
        elif split == Split.ALL:
            subfolders = ['train', 'val', 'test']

        self.info = []
        for subf in subfolders:
            for name in os.listdir(os.path.join(self.root, subf, 'images')):
                if not name.endswith('.jpg'):
                    continue
                name = name.split('.')[0]
                image_path = os.path.join(
                    self.root, subf, 'images', f'{name}.jpg')
                label_path = os.path.join(
                    self.root, subf, 'labels', f'{name}.png')
                landmark_path = os.path.join(
                    self.root, subf, 'landmarks', f'{name}.txt')
                assert os.path.exists(image_path)
                assert os.path.exists(label_path)
                assert os.path.exists(landmark_path)
                landmarks = [float(v) for v in open(
                    landmark_path, 'r').read().split()]
                assert landmarks[0] == 106 and len(landmarks) == 106*2+1
                landmarks = np.reshape(
                    np.array(landmarks[1:], np.float32), [106, 2])
                sample_name = f'{subf}.{name}'
                self.info.append(
                    {'image_path': image_path, 'label_path': label_path,
                     'landmarks': landmarks, 'sample_name': sample_name})
                if split == Split.TOY and len(self.info) >= 10:
                    break

    def __getitem__(self, index):
        info = self.info[index]
        image = cv2.imread(info['image_path'])[:, :, ::-1]
        label = cv2.imread(info['label_path'], cv2.IMREAD_GRAYSCALE)
        landmarks = info['landmarks']
        return {'image': image, 'label': label, 'landmarks': landmarks}

    def __len__(self):
        return len(self.info)

    def sample_name(self, index):
        return self.info[index]['sample_name']

    @property
    def label_names(self):
        return ['background', 'face_lr_rr', 'lb', 'rb', 'le', 're', 'nose', 'ul', 'im', 'll', 'hair']

    @staticmethod
    def draw_landmarks(im, landmarks, color, thickness=5, eye_radius=3):
        landmarks = landmarks.astype(np.int32)
        cv2.polylines(im, [landmarks[0:33]], False,
                      color, thickness, cv2.LINE_AA)
        cv2.polylines(im, [landmarks[33:42]], True,
                      color, thickness, cv2.LINE_AA)
        cv2.polylines(im, [landmarks[42:51]], True,
                      color, thickness, cv2.LINE_AA)
        cv2.polylines(im, [landmarks[51:55]], False,
                      color, thickness, cv2.LINE_AA)
        cv2.polylines(im, [landmarks[55:66]], False,
                      color, thickness, cv2.LINE_AA)
        cv2.polylines(im, [landmarks[66:74]], True,
                      color, thickness, cv2.LINE_AA)
        cv2.circle(im, (landmarks[74, 0], landmarks[74, 1]),
                   eye_radius, color, thickness, cv2.LINE_AA)
        cv2.polylines(im, [landmarks[75:83]], True,
                      color, thickness, cv2.LINE_AA)
        cv2.circle(im, (landmarks[83, 0], landmarks[83, 1]),
                   eye_radius, color, thickness, cv2.LINE_AA)
        cv2.polylines(im, [landmarks[84:96]], True,
                      color, thickness, cv2.LINE_AA)
        cv2.polylines(im, [landmarks[96:-2]], True,
                      color, thickness, cv2.LINE_AA)
        return im



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from typing import Union, Tuple
import cv2

from ..dataset import Dataset


class Wild(Dataset):
    """ Dataset based on wild images.
    """

    def __init__(self, root: str, suffix: Union[str, Tuple[str, ...]] = ('.jpg', '.png', '.jpeg')):
        self.root = root
        self.paths = []
        for r, _, files in os.walk(self.root):
            for name in files:
                if name.lower().endswith(suffix):
                    self.paths.append(os.path.join(r, name))

    def __len__(self):
        return len(self.paths)

    def __getitem__(self, index):
        p = self.paths[index]
        image = cv2.imread(p)[:, :, ::-1]
        return {'image': image}

    def sample_name(self, index):
        return self.paths[index].replace('\\', '.').replace('//', '.')



import setuptools

long_desc = open("README.md").read()
required = [
    'torch',
    'tqdm', 'pyyaml', 'numpy', 'matplotlib',
    'scikit-image', 'scikit-learn', 'opencv-python',
    'timm', 'xmltodict', 'tensorboard']

setuptools.setup(
    name="py-blueprint",
    version="0.0.1-1",
    author="Hao Yang",
    author_email="yanghao.alexis@foxmail.com",
    license="MIT",
    description="Configure Python to Yaml.",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url="https://github.com/YANG-H/BluePrint.git",
    packages=['blueprint', 'blueprint.ml', 'blueprint.ml.augmenters',
              'blueprint.ml.datasets'],
    key_words="Configuration, machine learning",
    install_requires=required,
    python_requires=">=3.7",
)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import collections
import torch


class BatchList:
    """A list representing a data batch. 

    It is not defined as a Sequence type to avoid ambiguity in collate/decollate.
    """

    def __init__(self, values) -> None:
        self.values = values

    def __repr__(self) -> str:
        return 'BatchList' + repr(self.values)

    def __len__(self) -> int:
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def __getitem__(self, index):
        return self.values[index]


class NonStackable:
    """A non-stackable object.

    A class for wraping objects that will not be stacked as tensors in `collate`.
    """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f'NonStackable[{self.value}]'


_np_str_obj_array_pattern = re.compile(r'[SaUO]')

_collate_err_msg_format = (
    "collate: batch must contain tensors, numpy arrays, numbers, "
    "dicts or lists; found {}")


def collate(data):
    """A replacement of `torch.utils.data.dataloader.default_collate`.

    Try collating a data batch into combined tensors.
    """

    elem = data[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum([x.numel() for x in data])
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return torch.stack(data, 0, out=out)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        elem = data[0]
        if elem_type.__name__ == 'ndarray':
            # array of string classes and object
            if _np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(
                    _collate_err_msg_format.format(elem.dtype))

            return collate([torch.as_tensor(b) for b in data])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(data)
    elif isinstance(elem, float):
        return torch.tensor(data, dtype=torch.float64)
    elif isinstance(elem, int):
        return torch.tensor(data)
    elif isinstance(elem, (str, bytes, NonStackable)):
        return BatchList(data)
    elif isinstance(elem, collections.abc.Mapping):
        return {key: collate([d[key] for d in data]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(collate(samples) for samples in zip(*data)))
    elif isinstance(elem, collections.abc.Sequence):
        transposed = zip(*data)
        return [collate(samples) for samples in transposed]


def _simplify_np_array(npa):
    if npa.shape == ():
        return npa.item()
    return npa


def decollate(data) -> BatchList:
    """The inverse process of `collate`.

    Try spliting the batched data back as numpy lists.
    """

    if isinstance(data, torch.Tensor):
        return BatchList([_simplify_np_array(e.detach().cpu().numpy())
                          for e in data.unbind(dim=0)])
    elif isinstance(data, BatchList):
        return data
    elif isinstance(data, (str, bytes, int, float, NonStackable)):
        raise TypeError(
            f'trying to decollate a non-batched type: {type(data)}')
    elif isinstance(data, collections.abc.Sequence):
        return BatchList(zip(*[decollate(e) for e in data]))
    elif isinstance(data, collections.abc.Mapping):
        data_dc = [(key, decollate(data[key])) for key in data]
        return BatchList([{key: val[ind] for key, val in data_dc}
                          for ind in range(len(data_dc[0][1]))])



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import numpy as np
import cv2
import functools
import torch
from typing import Dict, List

from . import data_path
from ..dataset import Dataset, Split


@functools.lru_cache()
def _cached_imread(fname, flags=None):
    return cv2.imread(fname, flags=flags)


class LFWA(Dataset):
    """ The LFWA dataset used by 

        Liu, Ziwei, Ping Luo, Xiaogang Wang, and Xiaoou Tang. 
        "Deep learning face attributes in the wild." In Proceedings 
        of the IEEE international conference on computer vision, 
        pp. 3730-3738. 2015.

    It can be downloaded from https://mmlab.ie.cuhk.edu.hk/projects/CelebA.html.

    """

    def __init__(self, split=Split.TRAIN, root=None, aligned=False):
        if root is None:
            root = data_path.get('LFWA')

        self.images_root = os.path.join(root, 'lfw')

        attr_data = torch.load(os.path.join(root, 'lfw_att_40.pth'))
        self.names = attr_data['name']
        self.labels = attr_data['label']

        index_data = torch.load(os.path.join(root, 'indices_train_test.pth'))
        if split == Split.TRAIN:
            self.indices = index_data["indices_img_train"].astype(np.long)-1
        elif split == Split.TEST:
            self.indices = index_data["indices_img_test"].astype(np.long)-1

        self.landmarks = torch.load(os.path.join(root, 'lfw_landmarks.pth'))

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, index):
        index = self.indices[index]
        image = cv2.cvtColor(_cached_imread(os.path.join(
            self.images_root, *self.names[index].split('\\'))), cv2.COLOR_BGR2RGB)
        labels = self.labels[index]
        landmarks = self.landmarks[index]
        return {
            'image': image,
            'attrs': labels,
            'face_5_pts': landmarks
        }

    def sample_name(self, index) -> str:
        index = self.indices[index]
        return self.names[index].replace('\\', '.')



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
from . import build, Context, get_user_config


def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument('job_cfg', type=str,
                        help='The experiment configuration file (.yaml) ')

    parser.add_argument('--exp_name', type=str, default=None)
    parser.add_argument('--blob_root', type=str, default=None)
    parser.add_argument('--copy2local', type=int, default=0)
    parser.add_argument('--debug_loss', type=int, default=1)
    parser.add_argument('--detect_anomaly', type=int, default=1)
    args, _other_args = parser.parse_known_args()

    other_args = dict()
    for idx in range(0, len(_other_args), 2):
        key = _other_args[idx]
        val = _other_args[idx+1]
        assert key.startswith('--')
        key = key[2:]
        other_args[key] = val

    exp_name = args.exp_name
    if exp_name is None:
        exp_name = get_user_config('exp_name', 'default_experiment')

    blob_root = args.blob_root
    if blob_root is None:
        blob_root = get_user_config('blob_root', './blob')

    print(f'====== RUNNING {args.job_cfg} ======')
    runnable = build(Context(exp_name=exp_name, blob_root=blob_root,
                             copy2local=args.copy2local == 1,
                             debug_loss=args.debug_loss == 1,
                             detect_anomaly=args.detect_anomaly == 1),
                     args.job_cfg, **other_args)
    runnable()


if __name__ == '__main__':
    _main()



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dataclasses import dataclass
from typing import Mapping, Union


from torch import nn
from copy import deepcopy


@dataclass
class EMAConfig:
    """EMA configuration.
    """
    network_name: str
    decay: float


def _interpret_ema_networks(networks: Mapping[str, Union[nn.Module, EMAConfig]]) -> Mapping[str, nn.Module]:
    outputs = dict()
    for name, net in networks.items():
        if isinstance(net, EMAConfig):
            orig_net = networks[net.network_name]
            assert isinstance(orig_net, nn.Module)
            new_net = deepcopy(orig_net)
            assert not hasattr(new_net, '_ema_cfg')
            new_net._ema_cfg = net
            net = new_net
        outputs[name] = net
    return outputs



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Dict, List, Union

import torch
import torch.distributed as dist
from torch.utils.data import DistributedSampler

from .dataset import Dataset, ConcatDataset


class DistributedRandomMixSampler(torch.utils.data.Sampler):
    """Distributed random mix sampler.

    Args:
        dataset (Dataset): a dataset concatenating multiple datasets.
        num_samples_each_thread (int): number of samples to iterate for each thread.
        seed (int): Seed for a local random sampler.
    """

    def __init__(self, dataset: Dataset,
                 num_samples_each_thread: Optional[int] = None,
                 num_replicas: Optional[int] = None,
                 rank: Optional[int] = None, seed: int = 0):
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError(
                    "Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError(
                    "Requires distributed package to be available")
            rank = dist.get_rank()
        if rank >= num_replicas or rank < 0:
            raise ValueError(
                "Invalid rank {}, rank should be in the interval"
                " [0, {}]".format(rank, num_replicas - 1))

        if not isinstance(dataset, ConcatDataset):
            dataset = ConcatDataset([dataset])

        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0

        if num_samples_each_thread is None:
            num_samples_each_thread = (
                min([len(ds) for ds in self.dataset.datasets]) +
                self.num_replicas - 1) // self.num_replicas

        self.num_samples_each_thread = num_samples_each_thread
        self.seed = seed

    def __iter__(self):
        indices_table = []
        acc_len = 0

        # deterministically shuffle based on epoch and seed
        g = torch.Generator()
        g.manual_seed(self.seed + self.epoch)
        for ds in self.dataset.datasets:
            indices_table.append(
                (torch.randperm(len(ds), generator=g)+acc_len))
            acc_len += len(ds)

        # subsample
        ds_id = self.rank % len(indices_table)
        index_ids = [0] * len(indices_table)
        for _ in range(self.num_samples_each_thread):
            index_id = index_ids[ds_id]
            sample_id = indices_table[ds_id][(
                index_id * self.num_replicas + self.rank) % len(indices_table[ds_id])]

            index_ids[ds_id] += 1
            ds_id = (ds_id + 1) % len(indices_table)
            yield sample_id

    def __len__(self) -> int:
        return self.num_samples_each_thread

    def set_epoch(self, epoch: int) -> None:
        self.epoch = epoch


UnifiedDistributedSampler = Union[DistributedSampler,
                                  DistributedRandomMixSampler]
